import random
import string
import time
from dataclasses import asdict
from typing import Optional, List

from aiogram.dispatcher import FSMContext
from aiogram.dispatcher.storage import FSMContextProxy

from .intent import Data, Intent

TASK_STACK = "__AIOGD_TASKS"
ID_SYMS = string.digits + string.ascii_letters


def new_id() -> int:
    return int(time.time()) % 100000000 + random.randint(0, 99) * 100000000


def id_to_str(int_id: int) -> str:
    if not int_id:
        return ID_SYMS[0]
    base = len(ID_SYMS)
    res = ""
    while int_id:
        int_id, mod = divmod(int_id, base)
        res += ID_SYMS[mod]
    return res


class DialogStack:
    def __init__(self, proxy: FSMContextProxy):
        self.proxy = proxy
        self.proxy.setdefault(TASK_STACK, [])

    @classmethod
    async def create(cls, state: FSMContext):
        proxy = await FSMContextProxy.create(state)
        return cls(proxy)

    def _stack(self) -> List:
        return self.proxy[TASK_STACK]

    def push(self, name: str, data: Data) -> Intent:
        # TODO check stack size overflow and data overflow
        intent = Intent(id_to_str(new_id()), name, data)
        self._stack().append(asdict(intent))
        return intent

    def current(self) -> Optional[Intent]:
        stack = self._stack()
        if not stack:
            return None
        return Intent(**stack[-1])

    def pop(self, intent: Optional[Intent] = None):
        stack = self._stack()
        if intent is None:
            stack.pop()
        else:
            stack.remove(intent)

    def clear(self):
        self._stack().clear()


if __name__ == '__main__':
    print(id_to_str(10000000000 - 1))
    stack = DialogStack({})
    print(stack.push("hello", 1))
    print(stack._stack())
    print(stack.proxy)
    print(stack.current())



from .base import Keyboard
from .button import Button, Url
from .checkbox import Checkbox
from .group import Group, Row, Column
from .select import Select, Radio, Multiselect
from .state import Back, Cancel, Next, Start, SwitchTo

__all__ = [
    "Keyboard",
    "Button", "Url",
    "Back", "Cancel", "Next", "Start", "SwitchTo",
    "Group", "Row", "Column",
    "Checkbox",
    "Select", "Radio", "Multiselect",
]



from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Dict, Optional, Any
from typing import Union, List

from aiogram import Bot
from aiogram.dispatcher.filters.state import State
from aiogram.types import Message, User, CallbackQuery, Chat

Data = Union[Dict, List, int, str, None]


@dataclass(frozen=True)
class Intent:
    id: str = field(compare=True)
    name: str = field(compare=False)
    data: Data = field(compare=False)


class Action(Enum):
    DONE = "DONE"
    START = "START"
    UPDATE = "UPDATE"
    SWITCH = "SWITCH"


@dataclass
class DialogUpdateEvent:
    bot: Bot
    from_user: User
    chat: Chat
    action: Action
    data: Any


@dataclass
class DialogStartEvent(DialogUpdateEvent):
    new_state: State
    reset_stack: bool


@dataclass
class DialogSwitchEvent(DialogUpdateEvent):
    new_state: State


ChatEvent = Union[CallbackQuery, Message, DialogUpdateEvent]



from typing import Union, Callable, Dict

from aiogram_dialog.manager.manager import DialogManager

Predicate = Callable[[Dict, "Whenable", DialogManager], bool]
WhenCondition = Union[str, Predicate, None]


def new_when_field(fieldname: str) -> Predicate:
    def when_field(data: Dict, widget: "Whenable", manager: DialogManager) -> bool:
        return bool(data.get(fieldname))

    return when_field


def true(data: Dict, widget: "Whenable", manager: DialogManager):
    return True


class Whenable:

    def __init__(self, when: WhenCondition = None):
        self.condition: Predicate
        if when is None:
            self.condition = true
        elif isinstance(when, str):
            self.condition = new_when_field(when)
        else:
            self.condition = when

    def is_(self, data, manager):
        return self.condition(data, self, manager)



from typing import Union, Sequence, Tuple, Callable

from .input import MessageHandlerFunc, BaseInput, MessageInput, TextInput
from .kbd import Keyboard, Group
from .text import Multi, Format, Text
from .widget_event import WidgetEventProcessor


def ensure_text(widget: Union[str, Text, Sequence[Text]]) -> Text:
    if isinstance(widget, str):
        return Format(widget)
    if isinstance(widget, Sequence):
        if len(widget) == 1:
            return widget[0]
        return Multi(*widget)
    return widget


def ensure_keyboard(widget: Union[Keyboard, Sequence[Keyboard]]) -> Keyboard:
    if isinstance(widget, Sequence):
        if len(widget) == 1:
            return widget[0]
        return Group(*widget)
    return widget


def ensure_input(
        widget: Union[MessageHandlerFunc, WidgetEventProcessor, BaseInput, Sequence[BaseInput]]
) -> BaseInput:
    if isinstance(widget, BaseInput):
        return widget
    elif isinstance(widget, Sequence):
        if len(widget) == 0:
            return MessageInput(None)
        elif len(widget) == 1:
            return widget[0]
        else:
            raise ValueError(f"Only 1 input supported, got {len(widget)}")
    else:
        return MessageInput(widget)


def ensure_widgets(
        widgets: Sequence[Union[str, Text, Keyboard, MessageHandlerFunc, BaseInput]]
) -> Tuple[Text, Keyboard, BaseInput]:
    texts = []
    keyboards = []
    inputs = []

    for w in widgets:
        if isinstance(w, (str, Text)):
            texts.append(ensure_text(w))
        elif isinstance(w, Keyboard):
            keyboards.append(ensure_keyboard(w))
        elif isinstance(w, (BaseInput, Callable)):
            inputs.append(ensure_input(w))
        else:
            raise TypeError("Invalid widget type: %s" % type(w))
    return ensure_text(texts), ensure_keyboard(keyboards), ensure_input(inputs)



from typing import Any, Optional, Type

from aiogram.dispatcher.filters.state import State, StatesGroup
from aiogram.dispatcher.storage import FSMContextProxy

GLOBAL_CONTEXT = "__AIOGD_G_CONTEXT"
DIALOG_CONTEXT = "__AIOGD_D_CONTEXT"
DIALOG_INTERNAL_CONTEXT = "__AIOGD_ID_CONTEXT"

LAST_MESSAGE_ID = "LAST_MESSAGE_ID"

FORBID = object()


def reset_dialog_contexts(proxy: FSMContextProxy):
    proxy.pop(DIALOG_CONTEXT, None)
    proxy.pop(DIALOG_INTERNAL_CONTEXT, None)


class DialogContext:
    def __init__(self, proxy: FSMContextProxy, dialog_id: str, states_group: Optional[Type[StatesGroup]]):
        self.proxy = proxy
        self.dialog_id = dialog_id
        self.states_group = states_group

    @property
    def state(self) -> State:
        if not self.states_group:
            raise ValueError("No states group in current context")
        state = self.proxy.state
        for s in self.states_group.states:
            if s.state == state:
                return s
        raise ValueError("Unknown state `%s`" % state)

    @state.setter
    def state(self, state: State):
        self.proxy.state = state.state

    @property
    def last_message_id(self) -> Optional[int]:
        gcontext = self.proxy.get(GLOBAL_CONTEXT) or {}
        return gcontext.get(LAST_MESSAGE_ID)

    @last_message_id.setter
    def last_message_id(self, last_message_id: int):
        self.proxy.setdefault(GLOBAL_CONTEXT, {})
        g_context = self.proxy[GLOBAL_CONTEXT]
        g_context[LAST_MESSAGE_ID] = last_message_id

    def set_data(self, key: str, value: Any, *, internal: bool = False):
        context_name = DIALOG_INTERNAL_CONTEXT if internal else DIALOG_CONTEXT
        self.proxy.setdefault(context_name, {})
        d_context = self.proxy[context_name]
        d_context.setdefault(self.dialog_id, {})
        d_context[self.dialog_id][key] = value

    def data(self, key, default=FORBID, *, internal: bool = False, ):
        context_name = DIALOG_INTERNAL_CONTEXT if internal else DIALOG_CONTEXT
        d_context = self.proxy.get(context_name) or {}
        dialog_data = d_context.get(self.dialog_id) or {}
        if default is FORBID:
            return dialog_data[key]
        else:
            return dialog_data.get(key, default)

    def clear(self):
        if DIALOG_CONTEXT in self.proxy:
            if self.dialog_id in self.proxy[DIALOG_CONTEXT]:
                del self.proxy[DIALOG_CONTEXT][self.dialog_id]
        if DIALOG_INTERNAL_CONTEXT in self.proxy:
            if self.dialog_id in self.proxy[DIALOG_INTERNAL_CONTEXT]:
                del self.proxy[DIALOG_INTERNAL_CONTEXT][self.dialog_id]



from typing import Dict

from .base import Text
from ..when import WhenCondition
from ...manager.manager import DialogManager
from jinja2 import Template


class Jinja(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        # TODO Environment
        # from file
        # custom filters
        super().__init__(when)
        self.template = Template(text, trim_blocks=True, lstrip_blocks=True,autoescape=True)

    async def _render_text(self, data: Dict, manager: DialogManager) -> str:
        return self.template.render(data)



from typing import Callable, Union, Dict

from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.when import Whenable, WhenCondition


class Text(Whenable):
    def __init__(self, when: WhenCondition = None):
        super().__init__(when)

    async def render_text(self, data: Dict, manager: DialogManager) -> str:
        if not self.is_(data, manager):
            return ""
        return await self._render_text(data, manager)

    async def _render_text(self, data, manager: DialogManager) -> str:
        raise NotImplementedError


class Const(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        super().__init__(when)
        self.text = text

    async def _render_text(self, data: Dict, manager: DialogManager) -> str:
        return self.text



from typing import Callable, TypeVar, Generic, Awaitable, Union

from aiogram.types import Message, ContentType

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import BaseInput

T = TypeVar("T")
TypeFactory = Callable[[str], T]
OnSuccess = Callable[[Message, "TextInput", DialogManager, T], Awaitable]
OnError = Callable[[Message, "TextInput", DialogManager], Awaitable]


class TextInput(BaseInput, Generic[T]):
    def __init__(self, id: str, type_factory: TypeFactory[T] = str,
                 on_success: Union[OnSuccess[T], WidgetEventProcessor, None] = None,
                 on_error: Union[OnError, WidgetEventProcessor, None] = None):
        super().__init__(id)
        self.type_factory = type_factory
        self.on_success = ensure_event_processor(on_success)
        self.on_error = ensure_event_processor(on_error)

    async def process_message(self, message: Message, dialog: Dialog, manager: DialogManager):
        if message.content_type != ContentType.TEXT:
            return False
        try:
            value = self.type_factory(message.text)
            await self.on_success.process_event(message, self, manager, value)
            manager.context.set_data(self.widget_id, message.text, internal=True)  # store original text
        except ValueError:
            await self.on_error.process_event(message, self, manager)

    def get_value(self, manager: DialogManager):
        return self.type_factory(manager.context.data(self.widget_id, internal=True))



from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Callable, Optional, Union, Dict, Any, List, Awaitable, Sequence

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.intent import ChatEvent
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.text import Text, Case
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import Keyboard

ItemIdGetter = Callable[[Any], Union[str, int]]
ItemsGetter = Callable[[Dict], Sequence]
OnItemStateChanged = Callable[[ChatEvent, "Select", DialogManager, str], Awaitable]
OnItemClick = Callable[[CallbackQuery, "Select", DialogManager, str], Awaitable]


def get_identity(items: Sequence) -> ItemsGetter:
    def identity(data) -> Sequence:
        return items

    return identity


class Select(Keyboard):
    def __init__(self, text: Text,
                 id: str,
                 item_id_getter: ItemIdGetter,
                 items: Union[str, Sequence],
                 on_click: Union[OnItemClick, WidgetEventProcessor, None] = None,
                 when: Union[str, Callable] = None):
        super().__init__(id, when)
        self.text = text
        self.widget_id = id
        self.on_click = ensure_event_processor(on_click)
        self.callback_data_prefix = id + ":"
        self.item_id_getter = item_id_getter
        if isinstance(items, str):
            self.items_getter = itemgetter(items)
        else:
            self.items_getter = get_identity(items)

    async def _render_keyboard(self, data: Dict, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        return [[
            await self._render_button(pos, item, data, manager)
            for pos, item in enumerate(self.items_getter(data))
        ]]

    async def _render_button(self, pos: int, item: Any, data: Dict, manager: DialogManager) -> InlineKeyboardButton:
        data = {"data": data, "item": item, "pos": pos + 1, "pos0": pos}
        return InlineKeyboardButton(
            text=await self.text.render_text(data, manager),
            callback_data=self.callback_data_prefix + str(self.item_id_getter(item))
        )

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager) -> bool:
        if not c.data.startswith(self.callback_data_prefix):
            return False
        item_id = c.data[len(self.callback_data_prefix):]
        await self.on_click.process_event(c, self, manager, item_id)
        return True


class StatefulSelect(Select, ABC):
    def __init__(self, checked_text: Text, unchecked_text: Text,
                 id: str, item_id_getter: ItemIdGetter,
                 items: Union[str, Sequence],
                 on_click: Union[OnItemClick, WidgetEventProcessor, None] = None,
                 on_state_changed: Union[OnItemStateChanged, WidgetEventProcessor, None] = None,
                 when: Union[str, Callable] = None):
        text = Case({True: checked_text, False: unchecked_text}, selector=self._is_text_checked)
        super().__init__(text, id, item_id_getter, items, self._process_click, when)
        self.on_item_click = ensure_event_processor(on_click)
        self.on_state_changed = ensure_event_processor(on_state_changed)

    async def _process_on_state_changed(self, event: ChatEvent, item_id: str, manager: DialogManager):
        if self.on_state_changed:
            await self.on_state_changed.process_event(event, self, manager, item_id)

    @abstractmethod
    def _is_text_checked(self, data: Dict, case: Case, manager: DialogManager) -> bool:
        raise NotImplementedError

    async def _process_click(self, c: CallbackQuery, select: Select, manager: DialogManager, item_id: str):
        if self.on_item_click:
            await self.on_item_click.process_event(c, select, manager, item_id)
        await self._on_click(c, select, manager, item_id)

    @abstractmethod
    async def _on_click(self, c: CallbackQuery, select: Select, manager: DialogManager, item_id: str):
        raise NotImplementedError


class Radio(StatefulSelect):
    def get_checked(self, manager: DialogManager) -> Optional[str]:
        return manager.context.data(self.widget_id, None, internal=True)

    async def set_checked(self, event: ChatEvent, item_id: Optional[str], manager: DialogManager):
        checked = self.get_checked(manager)
        manager.context.set_data(self.widget_id, item_id, internal=True)
        if checked != item_id:
            await self._process_on_state_changed(event, item_id, manager)

    def is_checked(self, item_id: Union[str, int], manager: DialogManager) -> bool:
        return str(item_id) == self.get_checked(manager)

    def _is_text_checked(self, data: Dict, case: Case, manager: DialogManager) -> bool:
        item_id = str(self.item_id_getter(data["item"]))
        return self.is_checked(item_id, manager)

    async def _on_click(self, c: CallbackQuery, select: Select, manager: DialogManager, item_id: str):
        print(self, item_id)
        await self.set_checked(c, item_id, manager)


class Multiselect(StatefulSelect):
    def __init__(self, checked_text: Text, unchecked_text: Text, id: str, item_id_getter: ItemIdGetter,
                 items: Union[str, Sequence], min_selected: int = 0, max_selected: int = 0,
                 on_click: Union[OnItemClick, WidgetEventProcessor, None] = None,
                 on_state_changed: Union[OnItemStateChanged, WidgetEventProcessor, None] = None,
                 when: Union[str, Callable] = None):
        super().__init__(checked_text, unchecked_text, id, item_id_getter, items, on_state_changed, when)
        self.min_selected = min_selected
        self.max_selected = max_selected

    def _is_text_checked(self, data: Dict, case: Case, manager: DialogManager) -> bool:
        item_id = str(self.item_id_getter(data["item"]))
        return self.is_checked(item_id, manager)

    def is_checked(self, item_id: Union[str, int], manager: DialogManager) -> bool:
        data: List = self.get_checked(manager)
        return str(item_id) in data

    def get_checked(self, manager: DialogManager) -> List[str]:
        return manager.context.data(self.widget_id, [], internal=True)

    async def reset_checked(self, event: ChatEvent, manager: DialogManager):
        manager.context.set_data(self.widget_id, [], internal=True)

    async def set_checked(self, event: ChatEvent, item_id: str, checked: bool, manager: DialogManager) -> None:
        data: List = self.get_checked(manager)
        if item_id in data:
            if not checked:
                if len(data) > self.min_selected:
                    data.remove(item_id)
                    await self._process_on_state_changed(event, item_id, manager)
        else:
            if checked:
                if self.max_selected == 0 or self.max_selected > len(data):
                    data.append(item_id)
                    await self._process_on_state_changed(event, item_id, manager)
        manager.context.set_data(self.widget_id, data, internal=True)

    async def _on_click(self, c: CallbackQuery, select: Select, manager: DialogManager, item_id: str):
        await self.set_checked(c, item_id, not self.is_checked(item_id, manager), manager)



from .dialog import Dialog
from .manager.protocols import DialogManager
from .manager.registry import DialogRegistry
from .manager.bg_manager import BgManagerProto as BgManager
from .window import Window
from .manager.intent import ChatEvent

__all__ = [
    "Dialog",
    "ChatEvent",
    "BgManager",
    "DialogManager",
    "DialogRegistry",
    "Window",
]



from typing import Any, Dict, Optional

from aiogram import Bot
from aiogram.dispatcher.filters.state import State
from aiogram.types import Message, CallbackQuery, Chat, User

from .intent import (
    Intent, Data, Action, DialogStartEvent, DialogSwitchEvent, DialogUpdateEvent, ChatEvent
)
from .protocols import DialogRegistryProto, BgManagerProto


def get_chat(event: ChatEvent) -> Chat:
    if isinstance(event, (Message, DialogUpdateEvent)):
        return event.chat
    elif isinstance(event, CallbackQuery):
        return event.message.chat


class BgManager(BgManagerProto):
    def __init__(
            self,
            user: User,
            chat: Chat,
            bot: Bot,
            registry: DialogRegistryProto,
            intent: Intent,
            current_state: State
    ):
        self.user = user
        self.chat = chat
        self.bot = bot
        self.registry = registry
        self.intent = intent
        self.current_state = current_state

    def current_intent(self) -> Intent:
        return self.intent

    def bg(self, user_id: Optional[int] = None, chat_id: Optional[int] = None) -> BgManagerProto:
        if user_id is not None:
            user = User(id=user_id)
        else:
            user = self.user
        if chat_id is not None:
            chat = Chat(id=chat_id)
        else:
            chat = self.chat

        return BgManager(
            user,
            chat,
            self.bot,
            self.registry,
            self.current_intent(),
            self.current_state,
        )

    async def done(self, result: Any = None):
        await self.registry.notify(DialogUpdateEvent(
            self.bot,
            self.user,
            self.chat,
            Action.DONE,
            result,
        ))

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        await self.registry.notify(DialogStartEvent(
            self.bot,
            self.user,
            self.chat,
            Action.START,
            data,
            state,
            reset_stack,
        ))

    async def switch_to(self, state: State):
        await self.registry.notify(DialogSwitchEvent(
            self.bot,
            self.user,
            self.chat,
            Action.SWITCH,
            {},
            state,
        ))

    async def update(self, data: Dict):
        await self.registry.notify(DialogUpdateEvent(
            self.bot,
            self.user,
            self.chat,
            Action.UPDATE,
            data,
        ))



from typing import Callable, Awaitable, List, Union

from aiogram.dispatcher.filters import ContentTypeFilter
from aiogram.types import Message

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.action import Actionable
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor

MessageHandlerFunc = Callable[[Message, Dialog, DialogManager], Awaitable]


class BaseInput(Actionable):
    async def process_message(self, m: Message, dialog: Dialog, manager: DialogManager) -> bool:
        raise NotImplementedError


class MessageInput(BaseInput):
    def __init__(self, func: Union[MessageHandlerFunc, WidgetEventProcessor, None],
                 content_types: List[str] = ContentTypeFilter.default):
        super().__init__()
        self.func = ensure_event_processor(func)
        self.filter = ContentTypeFilter(content_types)

    async def process_message(self, message: Message, dialog: Dialog, manager: DialogManager) -> bool:
        if not await self.filter.check(message):
            return False
        await self.func.process_event(message, dialog, manager)
        return True



from typing import List, Callable, Optional, Union, Dict, Awaitable, Any

from aiogram.types import InlineKeyboardButton, CallbackQuery

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.text import Text
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import Keyboard
from ..when import WhenCondition

OnClick = Callable[[CallbackQuery, "Button", DialogManager], Awaitable]


class Button(Keyboard):
    def __init__(self, text: Text, id: str,
                 on_click: Union[OnClick, WidgetEventProcessor, None] = None,
                 when: WhenCondition = None):
        super().__init__(id, when)
        self.text = text
        self.on_click = ensure_event_processor(on_click)

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager) -> bool:
        if c.data != self.widget_id:
            return False
        await self.on_click.process_event(c, self, manager)
        return True

    async def _render_keyboard(self, data: Dict, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        return [[
            InlineKeyboardButton(
                text=await self.text.render_text(data, manager),
                callback_data=self.widget_id
            )
        ]]


class Url(Keyboard):
    def __init__(self, text: Text, url: Text, id: Optional[str] = None, when: Union[str, Callable, None] = None):
        super().__init__(id, when)
        self.text = text
        self.url = url

    async def _render_keyboard(self, data: Dict, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        return [[
            InlineKeyboardButton(
                text=await self.text.render_text(data, manager),
                url=await self.url.render_text(data, manager)
            )
        ]]



from operator import itemgetter
from typing import Dict, Union, Sequence, Callable

from .base import Text
from ..when import WhenCondition

from aiogram_dialog.manager.manager import DialogManager


ItemsGetter = Callable[[Dict], Sequence]


def get_identity(items: Sequence) -> ItemsGetter:
    def identity(data) -> Sequence:
        return items
    return identity


class List(Text):
    def __init__(self, field: Text, items: Union[str, Sequence], sep: str = "\n", when: WhenCondition = None):
        super().__init__(when)
        self.field = field
        self.sep = sep
        if isinstance(items, str):
            self.items_getter = itemgetter(items)
        else:
            self.items_getter = get_identity(items)

    async def _render_text(self, data: Dict, manager: DialogManager) -> str:
        texts = [
            await self.field.render_text({"data": data, "item": item, "pos": pos + 1, "pos0": pos}, manager)
            for pos, item in enumerate(self.items_getter(data))
        ]
        return self.sep.join(filter(None, texts))



from typing import List, Optional

from aiogram.types import InlineKeyboardButton, CallbackQuery

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.action import Actionable
from aiogram_dialog.widgets.when import WhenCondition, Whenable


class Keyboard(Actionable, Whenable):
    def __init__(self, id: Optional[str] = None, when: WhenCondition = None):
        Actionable.__init__(self, id)
        Whenable.__init__(self, when)

    async def render_keyboard(self, data, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        if not self.is_(data, manager):
            return []
        return await self._render_keyboard(data, manager)

    async def _render_keyboard(self, data, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        raise NotImplementedError

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager) -> bool:
        return False



from typing import Callable, Optional, Any

from aiogram.dispatcher.filters.state import State
from aiogram.types import CallbackQuery

from aiogram_dialog.manager.intent import ChatEvent
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.text import Text, Const
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor
from .button import Button, OnClick
from ..when import WhenCondition


class EventProcessorButton(Button, WidgetEventProcessor):
    async def process_event(self, event: ChatEvent, source: Any, manager: DialogManager, *args, **kwargs):
        await self._on_click(event, self, manager)

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        raise NotImplementedError


class SwitchTo(EventProcessorButton):

    def __init__(self, text: Text, id: str,
                 state: State,
                 on_click: Optional[OnClick] = None,
                 when: WhenCondition = None):
        super().__init__(text, id, self._on_click, when)
        self.text = text
        self.user_on_click = on_click
        self.state = state

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        if self.user_on_click:
            await self.user_on_click(c, self, manager)
        await manager.dialog().switch_to(self.state, manager)


class Next(EventProcessorButton):
    def __init__(self, text: Text = Const("Next"), callback_data: str = "__next__",
                 on_click: Optional[Callable] = None,
                 when: WhenCondition = None):
        super().__init__(text, callback_data, self._on_click, when)
        self.text = text
        self.callback_data = callback_data
        self.user_on_click = on_click

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        if self.user_on_click:
            await self.user_on_click(c, self, manager)
        await manager.dialog().next(manager)


class Back(EventProcessorButton):
    def __init__(self, text: Text = Const("Back"), id: str = "__back__",
                 on_click: Optional[Callable] = None,
                 when: WhenCondition = None):
        super().__init__(text, id, self._on_click, when)
        self.text = text
        self.callback_data = id
        self.user_on_click = on_click

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        if self.user_on_click:
            await self.user_on_click(c, self, manager)
        await manager.dialog().back(manager)


class Cancel(EventProcessorButton):
    def __init__(self, text: Text = Const("Cancel"), id: str = "__cancel__",
                 on_click: Optional[Callable] = None,
                 when: WhenCondition = None):
        super().__init__(text, id, self._on_click, when)
        self.text = text
        self.user_on_click = on_click

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        if self.user_on_click:
            await self.user_on_click(c, self, manager)
        await manager.done()


class Start(EventProcessorButton):
    def __init__(self, text: Text, id: str,
                 state: State,
                 on_click: Optional[OnClick] = None,
                 reset_stack: bool = False,
                 when: WhenCondition = None):
        super().__init__(text, id, self._on_click, when)
        self.text = text
        self.user_on_click = on_click
        self.state = state
        self.reset_stack = reset_stack

    async def _on_click(self, c: CallbackQuery, button: Button, manager: DialogManager):
        if self.user_on_click:
            await self.user_on_click(c, self, manager)
        await manager.start(self.state, reset_stack=self.reset_stack)



from typing import Dict

from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.when import WhenCondition
from .base import Text


class Progress(Text):
    def __init__(self, field: str, width: int = 10, filled="🟥", empty="⬜", when: WhenCondition = None):
        super().__init__(when)
        self.field = field
        self.width = width
        self.filled = filled
        self.empty = empty

    async def _render_text(self, data: Dict, manager: DialogManager) -> str:
        percent = data.get(self.field)
        done = round((self.width * percent) / 100)
        rest = self.width - done

        return self.filled * done + self.empty * rest + f" {percent: 2.0f}%"



from typing import Optional, Any, Dict

from aiogram.dispatcher.filters.state import State
from aiogram.dispatcher.storage import FSMContextProxy
from aiogram.types import CallbackQuery, Chat, User
from aiogram.utils.exceptions import MessageNotModified

from .bg_manager import BgManager, get_chat
from .intent import Data, Intent, ChatEvent
from .protocols import DialogRegistryProto, DialogManager, BgManagerProto
from .stack import DialogStack
from ..data import DialogContext, reset_dialog_contexts


async def remove_kbd_safe(event: ChatEvent, proxy: FSMContextProxy):
    try:
        if isinstance(event, CallbackQuery):
            await event.message.edit_reply_markup()
        else:
            stub_context = DialogContext(proxy, "", None)
            last_message_id = stub_context.last_message_id
            if last_message_id:
                await event.bot.edit_message_reply_markup(event.chat.id, last_message_id)
    except MessageNotModified:
        pass  # nothing to modify


class IncorrectBackgroundError(RuntimeError):
    pass


class DialogManagerImpl(DialogManager):
    def __init__(
            self, event: ChatEvent, stack: DialogStack,
            proxy: FSMContextProxy, registry: DialogRegistryProto,
            data: Dict
    ):
        self.disabled = False
        self.proxy = proxy
        self.stack = stack
        self.registry = registry
        self.event = event
        self.data = data
        self.context = self.load_context()

    def check_disabled(self):
        if self.disabled:
            raise IncorrectBackgroundError(
                "Detected background access to dialog manager. "
                "Please use background manager available via `manager.bg()` "
                "method to access methods from background tasks"
            )

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        self.check_disabled()
        if reset_stack:
            await remove_kbd_safe(self.event, self.proxy)
            reset_dialog_contexts(self.proxy)
            self.stack.clear()
        dialog = self.registry.find_dialog(state)
        self.stack.push(state.state, data)
        self.context = self.load_context()
        await dialog.process_start(self, data, state)

    async def done(self, result: Any = None):
        self.check_disabled()
        await self.dialog().process_close(result, self)
        await self.mark_closed()
        intent = self.current_intent()
        if intent:
            self.proxy.state = intent.name
        else:
            self.proxy.state = None
        dialog = self.dialog()
        self.context = self.load_context()
        if dialog:
            await dialog.process_result(result, self)
            await dialog.show(self)
        else:
            await remove_kbd_safe(self.event, self.proxy)

    async def mark_closed(self):
        self.check_disabled()
        if self.context:
            self.context.clear()
        self.stack.pop()

    def current_intent(self) -> Intent:
        self.check_disabled()
        return self.stack.current()

    def dialog(self):
        self.check_disabled()
        current = self.current_intent()
        if not current:
            return None
        return self.registry.find_dialog(current.name)

    def load_context(self) -> Optional[DialogContext]:
        self.check_disabled()
        dialog = self.dialog()
        if not dialog:
            return None
        return DialogContext(self.proxy, dialog.states_group_name(), dialog.states_group())

    async def switch_to(self, state):
        self.check_disabled()
        self.context.state = state

    def bg(self, user_id: Optional[int] = None, chat_id: Optional[int] = None) -> BgManagerProto:
        if self.disabled:
            raise IncorrectBackgroundError(
                "Please call `manager.bg()` "
                "before starting background task"
            )
        self.check_disabled()

        if user_id is not None:
            user = User(id=user_id)
        else:
            user = self.event.from_user
        if chat_id is not None:
            chat = Chat(id=chat_id)
        else:
            chat = get_chat(self.event)

        return BgManager(
            user,
            chat,
            self.event.bot,
            self.registry,
            self.current_intent(),
            self.context.state,
        )

    async def close_manager(self) -> None:
        await self.proxy.save()
        self.disabled = True
        del self.proxy
        del self.stack
        del self.registry
        del self.event
        del self.data
        del self.context



from .base import Const, Text
from .format import Format
from .multi import Multi, Case
from .jinja import Jinja
from .list import List
from .progress import Progress

__all__ = [
    "Const", "Text",
    "Format",
    "Multi", "Case",
    "Jinja",
    "List",
    "Progress",
]



from typing import Optional, Any, Protocol, Union, Type, Dict

from aiogram import Dispatcher
from aiogram.dispatcher.filters.state import State, StatesGroup
from aiogram.dispatcher.storage import FSMContextProxy

from aiogram_dialog.data import DialogContext
from .intent import Intent, Data, DialogUpdateEvent, ChatEvent


class ManagedDialogProto(Protocol):
    def register(self, registry: "DialogRegistryProto", dp: Dispatcher, *args, **kwargs) -> None:
        pass

    def states_group_name(self) -> str:
        pass

    def states_group(self) -> Type[StatesGroup]:
        pass

    async def process_close(self, result: Any, manager: "DialogManager"):
        pass

    async def process_start(self, manager: "DialogManager", start_data: Any,
                            state: Optional[State] = None) -> None:
        pass

    async def show(self, manager: "DialogManager"):
        pass

    async def process_result(self, result: Any, manager: "DialogManager"):
        pass

    async def next(self, manager: "DialogManager"):
        pass

    async def back(self, manager: "DialogManager"):
        pass

    async def switch_to(self, state: State, manager: "DialogManager"):
        pass

    def find(self, widget_id) -> Optional[Any]:
        pass


class DialogRegistryProto(Protocol):
    def find_dialog(self, state: Union[State, str]) -> ManagedDialogProto:
        pass

    async def notify(self, event: DialogUpdateEvent) -> None:
        pass


class DialogManager(Protocol):
    context: Optional[DialogContext]
    proxy: FSMContextProxy
    event: ChatEvent
    data: Dict

    def dialog(self) -> ManagedDialogProto:
        pass

    def current_intent(self) -> Intent:
        pass

    async def close(self):
        pass

    async def done(self, result: Any = None):
        pass

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        pass

    async def switch_to(self, state):
        pass

    def bg(self, user_id: Optional[int] = None, chat_id: Optional[int] = None) -> "BgManagerProto":
        pass

    async def close_manager(self) -> None:
        pass


class BgManagerProto(Protocol):
    def current_intent(self) -> Intent:
        pass

    async def done(self, result: Any = None):
        pass

    async def start(self, state: State, data: Data = None, reset_stack: bool = False):
        pass

    async def switch_to(self, state: State):
        pass

    async def update(self, data: Dict):
        pass

    def bg(self, user_id: Optional[int] = None, chat_id: Optional[int] = None) -> "BgManagerProto":
        pass



from typing import Callable, Union, Dict, Any, Hashable

from .base import Text
from ..when import WhenCondition
from ...manager.manager import DialogManager


class Multi(Text):
    def __init__(self, *texts: Text, sep="\n", when: WhenCondition = None):
        super().__init__(when)
        self.texts = texts
        self.sep = sep

    async def _render_text(self, data, manager: DialogManager) -> str:
        texts = [
            await t.render_text(data, manager)
            for t in self.texts
        ]
        return self.sep.join(filter(None, texts))


Selector = Callable[[Dict, "Case", DialogManager], Hashable]


def new_case_field(fieldname: str) -> Selector:
    def case_field(data: Dict, widget: "Case", manager: DialogManager) -> Hashable:
        return data.get(fieldname)

    return case_field


class Case(Text):
    def __init__(self, texts: Dict[Any, Text], selector: Union[str, Selector], when: WhenCondition = None):
        super().__init__(when)
        self.texts = texts
        if isinstance(selector, str):
            self.selector = new_case_field(selector)
        else:
            self.selector = selector

    async def _render_text(self, data, manager: DialogManager) -> str:
        selection = self.selector(data, self, manager)
        return await self.texts[selection].render_text(data, manager)



from logging import getLogger

from .intent import DialogUpdateEvent, DialogStartEvent, Action, DialogSwitchEvent
from .manager import DialogManager

logger = getLogger(__name__)


async def handle_update(event: DialogUpdateEvent, dialog_manager: DialogManager):
    if isinstance(event, DialogStartEvent):
        await dialog_manager.start(state=event.new_state, data=event.data, reset_stack=event.reset_stack)
    elif isinstance(event, DialogSwitchEvent):
        await dialog_manager.switch_to(state=event.new_state)
        await dialog_manager.dialog().show(dialog_manager)
    elif event.action is Action.UPDATE:
        if not dialog_manager.context:
            logger.warning("No context found")
            return
        if event.data:
            for k, v in event.data.items():
                dialog_manager.context.set_data(k, v)
        await dialog_manager.dialog().show(dialog_manager)
    elif event.action is Action.DONE:
        await dialog_manager.done(result=event.data)



from abc import ABC, abstractmethod
from typing import Callable, Optional, Union, Dict, Awaitable, List

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.intent import ChatEvent
from aiogram_dialog.manager.manager import DialogManager
from aiogram_dialog.widgets.text import Text, Case
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor, ensure_event_processor
from .base import Keyboard

OnStateChanged = Callable[[ChatEvent, "Checkbox", DialogManager], Awaitable]


class BaseCheckbox(Keyboard, ABC):
    def __init__(self, checked_text: Text, unchecked_text: Text,
                 id: str,
                 on_click: Union[OnStateChanged, WidgetEventProcessor, None] = None,
                 on_state_changed: Union[OnStateChanged, WidgetEventProcessor, None] = None,
                 when: Union[str, Callable] = None):
        super().__init__(id, when)
        self.text = Case({True: checked_text, False: unchecked_text}, selector=self._is_text_checked)
        self.on_click = ensure_event_processor(on_click)
        self.on_state_changed = ensure_event_processor(on_state_changed)
        self._callback_data_prefix = f"{self.widget_id}:"

    async def _render_keyboard(self, data: Dict, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        checked = int(self.is_checked(data, manager))
        # store current checked status in callback data
        return [[
            InlineKeyboardButton(
                text=await self.text.render_text(data, manager),
                callback_data=f"{self._callback_data_prefix}{checked}"
            )
        ]]

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager) -> bool:
        if not c.data.startswith(f"{self._callback_data_prefix}"):
            return False
        # remove prefix and cast "0" as False, "1" as True
        checked = c.data[len(self._callback_data_prefix):] != "0"
        await self.on_click.process_event(c, self, manager)
        await self.set_checked(c, not checked, manager)
        return True

    def _is_text_checked(self, data: Dict, case: Case, manager: DialogManager) -> bool:
        return self.is_checked(data, manager)

    @abstractmethod
    def is_checked(self, data: Dict, manager: DialogManager) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def set_checked(self, event: ChatEvent, checked: bool, manager: DialogManager):
        raise NotImplementedError


class Checkbox(BaseCheckbox):
    def __init__(self, checked_text: Text, unchecked_text: Text, id: str,
                 on_state_changed: Optional[OnStateChanged] = None, default: bool = False,
                 when: Union[str, Callable] = None):
        super().__init__(checked_text, unchecked_text, id, on_state_changed, when)
        self.default = default

    def is_checked(self, data: Dict, manager: DialogManager) -> bool:
        return manager.context.data(self.widget_id, self.default, internal=True)

    async def set_checked(self, event: ChatEvent, checked: bool, manager: DialogManager):
        manager.context.set_data(self.widget_id, checked, internal=True)
        await self.on_state_changed.process_event(event, self, manager)



from typing import Dict

from .base import Text
from ..when import WhenCondition
from ...manager.manager import DialogManager


class Format(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        super().__init__(when)
        self.text = text

    async def _render_text(self, data: Dict, manager: DialogManager) -> str:
        return self.text.format_map(data)



  #!/usr/bin/env python3
# -*- coding: utf-8 -*-
from os import path

from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))


setup(
    name='aiogram_dialog',
    description='Mini-framework for dialogs on top of aiogram',
    version='0.11.1',
    url='https://github.com/tishka17/aiogram_dialog',
    author='A. Tikhonov',
    author_email='17@itishka.org',
    license='Apache2',
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ],
    packages=find_packages(include=['aiogram_dialog', 'aiogram_dialog.*']),
    install_requires=[
        'aiogram<3',
        'jinja2',
    ],
    python_requires=">=3.6",
)



import asyncio
from contextvars import copy_context
from functools import partial
from typing import Dict, Optional, Union

from aiogram import Dispatcher, Bot
from aiogram.dispatcher.filters.state import State
from aiogram.dispatcher.handler import Handler
from aiogram.dispatcher.middlewares import BaseMiddleware
from aiogram.dispatcher.storage import FSMContextProxy
from aiogram.types import User, Chat

from .intent import DialogUpdateEvent
from .manager import DialogManagerImpl
from .protocols import ManagedDialogProto, DialogRegistryProto, DialogManager
from .stack import DialogStack
from .update_handler import handle_update


class DialogRegistry(BaseMiddleware, DialogRegistryProto):
    dialogs: Dict[str, ManagedDialogProto]

    def __init__(self, dp: Dispatcher, dialogs: Optional[Dict[str, ManagedDialogProto]] = None):
        super().__init__()
        self.dp = dp
        if dialogs is None:
            dialogs = {}
        self.dialogs = dialogs
        self.update_handler = Handler(dp, middleware_key="aiogd_update")
        self._register_middleware()
        self.register_update_handler(handle_update, state="*")

    def register(self, dialog: ManagedDialogProto, *args, **kwargs):
        name = dialog.states_group_name()
        if name in self.dialogs:
            raise ValueError(f"StatesGroup `{name}` is already used")
        self.dialogs[name] = dialog
        dialog.register(self, self.dp, *args, **kwargs)

    def _register_middleware(self):
        self.dp.setup_middleware(self)

    def find_dialog(self, state: Union[str, State]) -> ManagedDialogProto:
        if isinstance(state, str):
            group, *_ = state.partition(":")
        else:
            group = state.group.__full_group_name__
        return self.dialogs[group]

    async def on_pre_process_message(self, event, data: dict):
        proxy = await FSMContextProxy.create(
            self.dp.current_state()  # there is no state in data at this moment
        )
        manager = DialogManagerImpl(
            event,
            DialogStack(proxy),
            proxy,
            self,
            data,
        )
        data["dialog_manager"] = manager

    on_pre_process_callback_query = on_pre_process_message
    on_pre_process_aiogd_update = on_pre_process_message

    async def on_post_process_message(self, _, result, data: dict):
        manager: DialogManager = data.pop("dialog_manager")
        await manager.close_manager()

    on_post_process_callback_query = on_post_process_message
    on_post_process_aiogd_update = on_post_process_message

    def register_update_handler(self, callback, *custom_filters, run_task=None, **kwargs) -> None:
        filters_set = self.dp.filters_factory.resolve(
            self.update_handler, *custom_filters, **kwargs
        )
        self.update_handler.register(self.dp._wrap_async_task(callback, run_task), filters_set)

    async def notify(self, event: DialogUpdateEvent) -> None:
        callback = lambda: asyncio.create_task(self._process_update(event))

        asyncio.get_running_loop().call_soon(
            callback,
            context=copy_context()
        )

    async def _process_update(self, event: DialogUpdateEvent):
        Bot.set_current(event.bot)
        User.set_current(event.from_user)
        Chat.set_current(event.chat)
        await self.update_handler.notify(event)



import re
from typing import Optional

ID_PATTERN = re.compile("^[a-zA-Z0-9_.]+$")


class Actionable:
    def __init__(self, id: Optional[str] = None):
        if id and not ID_PATTERN.match(id):
            raise ValueError(f"Invalid widget id: {id}")
        self.widget_id = id

    def find(self, widget_id):
        if self.widget_id is not None and self.widget_id == widget_id:
            return self
        return None



from itertools import chain
from typing import List, Dict, Optional

from aiogram.types import InlineKeyboardButton, CallbackQuery

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.manager.manager import DialogManager
from .base import Keyboard
from ..when import WhenCondition


class Group(Keyboard):
    def __init__(self, *buttons: Keyboard, id: Optional[str] = None, keep_rows: bool = True, width: int = 0,
                 when: WhenCondition = None):
        super().__init__(id, when)
        self.buttons = buttons
        self.keep_rows = keep_rows
        self.width = width

    def find(self, widget_id):
        widget = super(Group, self).find(widget_id)
        if widget:
            return widget
        for btn in self.buttons:
            widget = btn.find(widget_id)
            if widget:
                return widget
        return None

    async def _render_keyboard(self, data: Dict, manager: DialogManager) -> List[List[InlineKeyboardButton]]:
        kbd: List[List[InlineKeyboardButton]] = []
        for b in self.buttons:
            b_kbd = await b.render_keyboard(data, manager)
            if self.keep_rows or not kbd:
                kbd += b_kbd
            else:
                kbd[0].extend(chain.from_iterable(b_kbd))
        if not self.keep_rows and self.width:
            kbd = self._wrap_kbd(kbd[0])
        return kbd

    def _wrap_kbd(self, kbd: List[InlineKeyboardButton]) -> List[List[InlineKeyboardButton]]:
        res: List[List[InlineKeyboardButton]] = []
        row: List[InlineKeyboardButton] = []
        for b in kbd:
            row.append(b)
            if len(row) >= self.width:
                res.append(row)
                row = []
        if row:
            res.append(row)
        return res

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager) -> bool:
        for b in self.buttons:
            if await b.process_callback(c, dialog, manager):
                return True
        return False


class Row(Group):
    def __init__(self, *buttons: Keyboard, id: Optional[str] = None, when: WhenCondition = None):
        super().__init__(*buttons, id=id, keep_rows=False, when=when)


class Column(Group):
    def __init__(self, *buttons: Keyboard, id: Optional[str] = None, when: WhenCondition = None):
        super().__init__(*buttons, id=id, keep_rows=False, when=when, width=1)



from logging import getLogger
from typing import Dict, Optional, Union

from aiogram.dispatcher.filters.state import State
from aiogram.types import InlineKeyboardMarkup, Message, CallbackQuery, ParseMode
from aiogram.utils.exceptions import MessageNotModified

from aiogram_dialog.manager.protocols import DialogManager
from .dialog import Dialog, DialogWindowProto, DataGetter
from .manager.intent import DialogUpdateEvent
from .widgets.action import Actionable
from .widgets.input import BaseInput, MessageHandlerFunc
from .widgets.kbd import Keyboard
from .widgets.text import Text
from .widgets.utils import ensure_widgets

logger = getLogger(__name__)


class Window(DialogWindowProto):
    def __init__(self,
                 *widgets: Union[str, Text, Keyboard, MessageHandlerFunc, BaseInput],
                 state: State,
                 getter: DataGetter = None,
                 parse_mode: ParseMode = None):
        self.text, self.keyboard, self.on_message = ensure_widgets(widgets)
        self.getter = getter
        self.state = state
        self.parse_mode = parse_mode

    async def render_text(self, data: Dict, manager: DialogManager) -> str:
        return await self.text.render_text(data, manager)

    async def render_kbd(self, data: Dict, manager: DialogManager) -> InlineKeyboardMarkup:
        return InlineKeyboardMarkup(
            inline_keyboard=await self.keyboard.render_keyboard(data, manager)
        )

    async def load_data(self, dialog: "Dialog", manager: DialogManager) -> Dict:
        if not self.getter:
            return {}
        return await self.getter(**manager.data)

    async def process_message(self, message: Message, dialog: Dialog, manager: DialogManager):
        if self.on_message:
            await self.on_message.process_message(message, dialog, manager)

    async def process_callback(self, c: CallbackQuery, dialog: Dialog, manager: DialogManager):
        if self.keyboard:
            await self.keyboard.process_callback(c, dialog, manager)

    async def show(self, dialog: Dialog, manager: DialogManager) -> Message:
        logger.debug("Show window: %s", self)
        current_data = await self.load_data(dialog, manager)
        text = await self.render_text(current_data, manager)
        kbd = await self.render_kbd(current_data, manager)
        event = manager.event
        context = manager.context
        if isinstance(event, CallbackQuery):
            if text == event.message.text:
                if kbd != event.message.reply_markup:
                    return await event.message.edit_reply_markup(reply_markup=kbd)
                else:
                    return event.message
            else:
                return await event.message.edit_text(
                    text=text, reply_markup=kbd, parse_mode=self.parse_mode
                )
        elif isinstance(event, DialogUpdateEvent):
            if context and context.last_message_id:
                try:
                    return await event.bot.edit_message_text(
                        message_id=context.last_message_id, chat_id=event.chat.id,
                        text=text, reply_markup=kbd, parse_mode=self.parse_mode
                    )
                except MessageNotModified:
                    pass  # nothing to update
        else:
            if context and context.last_message_id:
                try:
                    await manager.event.bot.edit_message_reply_markup(
                        message_id=context.last_message_id, chat_id=manager.event.chat.id
                    )
                except MessageNotModified:
                    pass  # nothing to remove
            return await manager.event.bot.send_message(
                chat_id=event.chat.id, text=text, reply_markup=kbd, parse_mode=self.parse_mode
            )

    def get_state(self) -> State:
        return self.state

    def find(self, widget_id) -> Optional[Actionable]:
        if self.keyboard:
            res = self.keyboard.find(widget_id)
            if res:
                return res
        if self.on_message:
            return self.on_message.find(widget_id)
        return None

    def __repr__(self):
        return f"<{self.__class__.__qualname__}({self.state})>"



from typing import Union, Callable, Any

from aiogram_dialog.manager.intent import ChatEvent
from aiogram_dialog.manager.protocols import DialogManager


class WidgetEventProcessor:
    async def process_event(self, event: ChatEvent, source: Any, manager: DialogManager, *args, **kwargs):
        raise NotImplementedError


class SimpleEventProcessor(WidgetEventProcessor):
    def __init__(self, callback: Callable):
        self.callback = callback

    async def process_event(self, event: ChatEvent, source: Any, manager: DialogManager, *args, **kwargs):
        if self.callback:
            await self.callback(event, source, manager, *args, **kwargs)


def ensure_event_processor(processor: Union[Callable, WidgetEventProcessor, None]) -> WidgetEventProcessor:
    if isinstance(processor, WidgetEventProcessor):
        return processor
    else:
        return SimpleEventProcessor(processor)



from logging import getLogger
from typing import Dict, Callable, Awaitable, List, Union, Any, Optional, Type, TypeVar
from typing import Protocol

from aiogram import Dispatcher
from aiogram.dispatcher.filters.state import State, StatesGroup
from aiogram.types import InlineKeyboardMarkup, Message, CallbackQuery, ContentTypes

from aiogram_dialog.manager.protocols import DialogRegistryProto, ManagedDialogProto, DialogManager
from aiogram_dialog.widgets.action import Actionable

logger = getLogger(__name__)
DIALOG_CONTEXT = "DIALOG_CONTEXT"
DataGetter = Callable[..., Awaitable[Dict]]

ChatEvent = Union[CallbackQuery, Message]
OnDialogEvent = Callable[[Any, DialogManager], Awaitable]
W = TypeVar("W", bound=Awaitable)


class DialogWindowProto(Protocol):
    async def render_text(self, data: Dict, manager: DialogManager) -> str:
        raise NotImplementedError

    async def render_kbd(self, data: Dict, manager: DialogManager) -> InlineKeyboardMarkup:
        raise NotImplementedError

    async def load_data(self, dialog: "Dialog", manager: DialogManager) -> Dict:
        raise NotImplementedError

    async def process_message(self, m: Message, dialog: "Dialog", manager: DialogManager):
        raise NotImplementedError

    async def process_callback(self, c: CallbackQuery, dialog: "Dialog", manager: DialogManager):
        raise NotImplementedError

    async def show(self, dialog: "Dialog", manager: DialogManager) -> Message:
        raise NotImplementedError

    def get_state(self) -> State:
        raise NotImplementedError

    def find(self, widget_id) -> Optional[Actionable]:
        raise NotImplementedError


class Dialog(ManagedDialogProto):
    def __init__(
            self,
            *windows: DialogWindowProto,
            on_start: Optional[OnDialogEvent] = None,
            on_close: Optional[OnDialogEvent] = None,
            on_process_result: Optional[OnDialogEvent] = None
    ):
        self._states_group = windows[0].get_state().group
        self.states: List[State] = []
        for w in windows:
            if w.get_state().group != self._states_group:
                raise ValueError("All windows must be attached to same StatesGroup")
            state = w.get_state()
            if state in self.states:
                raise ValueError(f"Multiple windows with state {state}")
            self.states.append(state)
        self.windows: Dict[State, DialogWindowProto] = dict(zip(self.states, windows))
        self.on_start = on_start
        self.on_close = on_close
        self.on_process_result = on_process_result

    async def next(self, manager: DialogManager):
        if not manager.context:
            raise ValueError("No context")
        new_state = self.states[self.states.index(manager.context.state) + 1]
        await self.switch_to(new_state, manager)

    async def back(self, manager: DialogManager):
        if not manager.context:
            raise ValueError("No context")
        new_state = self.states[self.states.index(manager.context.state) - 1]
        await self.switch_to(new_state, manager)

    async def process_start(self, manager: DialogManager, start_data: Any, state: Optional[State] = None) -> None:
        if state is None:
            state = self.states[0]
        logger.debug("Dialog start: %s (%s)", state, self)
        await self.switch_to(state, manager)
        await self._process_callback(self.on_start, start_data, manager)
        await self.show(manager)

    async def _process_callback(self, callback: Optional[OnDialogEvent], data: Any, manager: DialogManager):
        if callback:
            await callback(data, manager)

    async def switch_to(self, state: State, manager: DialogManager):
        if state.group != self.states_group():
            raise ValueError("Cannot switch from %s to another states group %s" % (state.group, self.states_group()))
        await manager.switch_to(state)

    async def _current_window(self, manager: DialogManager) -> DialogWindowProto:
        return self.windows[manager.context.state]

    async def show(self, manager: DialogManager) -> None:
        logger.debug("Dialog show (%s)", self)
        window = await self._current_window(manager)
        message = await window.show(self, manager)
        manager.context.last_message_id = message.message_id

    async def _message_handler(self, m: Message, dialog_manager: DialogManager):
        intent = dialog_manager.current_intent()
        window = await self._current_window(dialog_manager)
        await window.process_message(m, self, dialog_manager)
        if dialog_manager.current_intent() == intent:  # no new dialog started
            await self.show(dialog_manager)

    async def _callback_handler(self, c: CallbackQuery, dialog_manager: DialogManager):
        intent = dialog_manager.current_intent()
        window = await self._current_window(dialog_manager)
        await window.process_callback(c, self, dialog_manager)
        if dialog_manager.current_intent() == intent:  # no new dialog started
            await self.show(dialog_manager)
        await c.answer()

    async def _update_handler(self, event: ChatEvent, dialog_manager: DialogManager):
        await self.show(dialog_manager)

    def register(self, registry: DialogRegistryProto, dp: Dispatcher, *args, **filters) -> None:
        dp.register_callback_query_handler(self._callback_handler, state=self.states, **filters)
        if "content_types" not in filters:
            filters["content_types"] = ContentTypes.ANY
        dp.register_message_handler(self._message_handler, state=self.states, *args, **filters)

    def states_group(self) -> Type[StatesGroup]:
        return self._states_group

    def states_group_name(self) -> str:
        return self._states_group.__full_group_name__

    async def process_result(self, result: Any, manager: DialogManager):
        await self._process_callback(self.on_process_result, result, manager)

    async def process_close(self, result: Any, manager: DialogManager):
        await self._process_callback(self.on_close, result, manager)

    def find(self, widget_id) -> Optional[W]:
        for w in self.windows.values():
            widget = w.find(widget_id)
            if widget:
                return widget
        return

    def __repr__(self):
        return f"<{self.__class__.__qualname__}({self.states_group()})>"
